#!/usr/bin/env python
# coding: utf-8

"""
main script

@author: Moritz Schubert
"""

# standard library
import os
import pickle

# external
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import pandas as pd
from tqdm import tqdm

# local
from bcibo_class import BCIBO_Model


def round_decimals_up(number:float, decimals:int=2):
    """
    Returns a value rounded up to a specific number of decimal places.
    
    Code copied from https://kodify.net/python/math/round-decimals/
    License terms: https://kodify.net/terms-privacy/#copyright-of-code-examples
    """
    if not isinstance(decimals, int):
        raise TypeError("decimal places must be an integer")
    elif decimals < 0:
        raise ValueError("decimal places has to be 0 or more")
    elif decimals == 0:
        return np.ceil(number)

    factor = 10 ** decimals
    return np.ceil(number * factor) / factor


def diff_between_dfs(df1, name1, df2, name2):
    
    # difference between the models for Samad et al.'s parameter choice
    df_diff = df1[['sigma_exponent', 'distance']]
    
    pH1_df1 = df1['p(H1)_mean']
    pH1_df2 = df2['p(H1)_mean']
    diff = pH1_df1 - pH1_df2
    df_diff['diff'] = diff.values
    df_diff['diff_round3up'] = round_decimals_up(diff.values, 3)
    df_diff = df_diff.reset_index(drop=True)
    
    df_diff.to_csv(os.path.join(outputDir, f'diff_{name1}-{name2}.csv'),
                   index=False)
    
    return df_diff


def distance_simulation(fname,
                        outputDir: str = 'output',
                        exponents = range(0, 35+1, 5),
                        p1: float = .5,
                        seed=None,
                        truncated: bool = False,
                        display_p: bool = False,):
    
    distances_mm = range(160, 360+1, 20) # distances in cm
    positionFH = 320 # FH: flesh hand
    
    # creates 2D array with hierarchical structure, sigma_exponent is the
    #  higher level
    df = pd.DataFrame({'sigma_exponent': np.repeat(exponents, len(distances_mm)),
                       'distance': np.tile(distances_mm, len(exponents))})

    for distance_mm in tqdm(distances_mm):
        positionRH = positionFH - distance_mm # RH: rubber hand
        
        for exponent in exponents:
            sigma = 10 ** exponent
            model = BCIBO_Model(xv_mu = positionRH,
                                sigma = sigma,
                                truncated=truncated,
                                c1 = p1,
                                seed=seed)
            
            df.loc[(df.sigma_exponent == exponent) & (df.distance == distance_mm),
                   'p(H1)_mean'] = model.p1
            df.loc[(df.sigma_exponent == exponent) & (df.distance == distance_mm),
                   'p(H1)_mean_round3'] = round(model.p1, 3)
            df.loc[(df.sigma_exponent == exponent) & (df.distance == distance_mm),
                    'p(H1)_sem'] = model.sem1
            # round up, so that we can say that all values are below a certain
            #  threshold
            df.loc[(df.sigma_exponent == exponent) & (df.distance == distance_mm),
                    'p(H1)_sem_round3up'] = round_decimals_up(model.sem1, 3)
    
    df.to_csv(os.path.join(outputDir, f'{fname}.csv'))
    
    fig, ax = plot_distance(fname, outputDir, p1=p1, df = df, display_p=display_p)
    
    fig.savefig(
        os.path.join(outputDir, f"{fname}.png"),
        dpi=200, bbox_inches='tight',
        )
    
    fig.savefig(
        os.path.join('../reports/seatedvr/journal/images', f'{fname}.pdf'),
        bbox_inches='tight')

    fig.savefig(
        os.path.join(outputDir, f"{fname}.pdf"),
        bbox_inches='tight',
        )
    
    return df


def plot_distance(
        fname: str,
        outputDir: str = 'output/',
        df: pd.core.frame.DataFrame = pd.DataFrame(),
        p1: float = .5,
        display_p: bool = False
) -> None:
    """
    analyse and visualise model for different distancces between real and rubber hand

    Returns
    -------
    None
        only side effects through output of plots

    """
    x_distance = df['distance'].unique()
    exponents =  df['sigma_exponent'].unique()
    
    # distance plotting
    fig, ax = plt.subplots(figsize=[5.5, 5.1])
    for i, exponent in enumerate(exponents):
        y_prob = df.loc[df['sigma_exponent']==exponent, 'p(H1)_mean'].values
        exponent_str = str(round(exponent))
        ax.plot(
            x_distance, y_prob, "x-", label="$10^{" + exponent_str + "}$",
            zorder=9 - i
        )
    
    ax.set_xticks(x_distance)
    ax.set_xlabel("distance between real and rubber hand (mm)")
    ax.set_ylabel("posterior probability of $C_1$")     
    if display_p:
        ax.set_title(f'$p(C_1) = {p1}$')
    
    #plt.tight_layout()
    
    ax.grid(alpha=0.5, zorder=0)
    fig.legend(title="$\sigma$ [mm/ms]", loc="upper right",
               bbox_to_anchor=(0.85, 0.85))
    
    return fig, ax



#%% main
if __name__ == "__main__":
    # use random seed for data sampling
    seed = None
    
    # uncomment to use the seed that recreates Figure 3 in 
    #  Schubert and Endres (in press)
    file = open('randomseed.pkl', 'rb')
    seed = pickle.load(file)
    file.close()
    
    outputDir = 'output'
    
    if not os.path.exists(outputDir):
        os.mkdir(outputDir)
    
    # # Figure 3 (right)
    df_original = distance_simulation('original',
                                      seed=seed)
    
    # Figure 4 (left)
    df_p50 = distance_simulation('50percent',
                                  seed=seed, display_p=True)
    
    # Figure 4 (right)
    df_p99 = distance_simulation('99percent',
                                  seed=seed, p1=.99, display_p=True)
    
    # # sigma values for priors on human scale
    # df_humanscale = distance_simulation('humanScale',
    #                                     seed=seed, exponents = [4])
    
    # # hyperprior=.7
    # df_p70 = distance_simulation('70percent',
    #                               seed=seed, p1=.7, display_p=True)
    
    # # hyperprior=1-10**-16
    # df_nearly_p100 = distance_simulation('closeTo100percent',
    #                                       seed=seed,
    #                                       p1=1-10**-16, display_p=True)
    
    # # truncated model
    # df_truncated = distance_simulation('truncated',
    #                                     truncated=True, seed=seed)
    
    # # difference between original and truncated model
    # diff_truncated = diff_between_dfs(df_original, 'original',
    #                                   df_truncated, 'truncated')
    
    # # difference between original and hyperprior=-.99 model
    # diff_p99 = diff_between_dfs(df_original, 'original',
    #                             df_p99, '99percent')
