#!/usr/bin/env python
# coding: utf-8

"""
main script

@author: Moritz Schubert
"""

# standard library
import os
import pickle
from typing import Literal

# external
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm import tqdm

# local
from bcibo_class_truncateComOnce import BCIBO_Model


def round_decimals_up(number:float, decimals:int=2):
    """
    Returns a value rounded up to a specific number of decimal places.
    
    Code copied from https://kodify.net/python/math/round-decimals/
    License terms: https://kodify.net/terms-privacy/#copyright-of-code-examples
    """
    if not isinstance(decimals, int):
        raise TypeError("decimal places must be an integer")
    elif decimals < 0:
        raise ValueError("decimal places has to be 0 or more")
    elif decimals == 0:
        return np.ceil(number)

    factor = 10 ** decimals
    return np.ceil(number * factor) / factor


def diff_original_truncated(df_original, df_truncated):
    
    # difference between the models for Samad et al.'s parameter choice
    df_diff = pd.DataFrame()
    for col in ['sigma_exponent', 'distance']:
        df_diff[col] = df_original[col]
    
    pH1_original = df_original['p(H1)_mean']
    pH1_truncated = df_truncated['p(H1)_mean']
    diff = pH1_truncated - pH1_original
    df_diff['diff'] = diff.values
    df_diff['diff_round3up'] = round_decimals_up(diff.values, 3)
    df_diff = df_diff.reset_index(drop=True)
    
    df_diff.to_csv(os.path.join(outputDir, 'diff_truncated-original.csv'),
               index=False)
    
    return df_diff


def distance_simulation(outputDir,
                        seed=None,
                        truncated:bool=False):
    
    distances_cm = range(16, 36+1, 2) # distances in cm
    exponents = range(0, 35+1, 5)
    positionFH = 320 # FH: flesh hand
    model_variant = 'double-truncated' if truncated else 'original'
    
    # creates 2D array with hierarchical structure, sigma_exponent is the
    #  higher level
    df = pd.DataFrame({'sigma_exponent': np.repeat(exponents, len(distances_cm)),
                       'distance': np.tile(distances_cm, len(exponents))})

    for distance_cm in tqdm(distances_cm):
        distance_mm = distance_cm * 10  # convert to mm, FH: flesh hand
        positionRH = positionFH - distance_mm # RH: rubber hand
        
        for exponent in exponents:
            sigma = 10 ** exponent
            model = BCIBO_Model(xv_mu = positionRH,
                                sigma = sigma,
                                truncated=truncated,
                                seed=seed)
            
            df.loc[(df.sigma_exponent == exponent) & (df.distance == distance_cm),
                   'p(H1)_mean'] = model.p1
            df.loc[(df.sigma_exponent == exponent) & (df.distance == distance_cm),
                   'p(H1)_mean_round3'] = round(model.p1, 3)
            df.loc[(df.sigma_exponent == exponent) & (df.distance == distance_cm),
                    'p(H1)_sem'] = model.sem1
            # round up, so that we can say that all values are below a certain
            #  threshold
            df.loc[(df.sigma_exponent == exponent) & (df.distance == distance_cm),
                    'p(H1)_sem_round3up'] = round_decimals_up(model.sem1, 3)
    
    df.to_csv(os.path.join(outputDir, f'distanceSimulation_{model_variant}.csv'))
    
    plot_distance(model_variant, outputDir, df = df)
    
    return df


def plot_distance(
        model_variant: Literal['original','double-truncated'],
        outputDir: str = 'output/',
        df: pd.core.frame.DataFrame = pd.DataFrame(),
        p1: float = .5,
) -> None:
    """
    analyse and visualise model for different distancces between real and rubber hand

    Returns
    -------
    None
        only side effects through output of plots

    """
    x_distance = df['distance'].unique()
    exponents =  df['sigma_exponent'].unique()
    
    # distance plotting
    fig, ax = plt.subplots(figsize=[6, 5.55])
    for i, exponent in enumerate(exponents):
        y_prob = df.loc[df['sigma_exponent']==exponent, 'p(H1)_mean'].values
        exponent_str = str(round(exponent))
        ax.plot(
            x_distance, y_prob, "x-", label="$10^{" + exponent_str + "}$",
            zorder=9 - i
        )
    
    ax.set_xticks(x_distance)
    ax.set_xlabel("distance between real and rubber hand [cm]")
    ax.set_ylabel("posterior probability of $C_1$")     
    ax.set_title(', '.join([model_variant, str(p1)]))
    
    ax.grid(alpha=0.5, zorder=0)
    fig.legend(title="$\sigma$ of prior", loc="upper right", bbox_to_anchor=(0.85, 0.85))
    
    return fig, ax


#%% main
if __name__ == "__main__":
    # use random seed for data sampling
    seed = None
    
    # uncomment to use the seed that recreates Figure 3 in 
    #  Schubert and Endres (in press)
    file = open('randomseed.pkl', 'rb')
    seed = pickle.load(file)
    file.close()
    
    outputDir = 'output'
    
    if not os.path.exists(outputDir):
        os.mkdir(outputDir)

    #df_original = distance_simulation(outputDir, truncated=False, seed=seed)
    df_truncated = distance_simulation(outputDir, truncated=True, seed=seed)
    
    #df_diff = diff_original_truncated(df_original, df_truncated)
    
    